package PureMessage::Scheduler;
use strict;
use Carp;
use HTTP::Date qw(time2iso);
use PureMessage::Version qw($PREFIX);
use PureMessage::Scheduler::Config;
use PureMessage::Scheduler::Event;
use PureMessage::Scheduler::Notify;
use PureMessage::Logger;
use PureMessage::Config qw(expand_filename %config);
use ActiveState::File::Atomic;
use Net::Domain qw(hostfqdn);

sub new {
    my $class = shift;
    $class = ref($class) || $class;
    my %param = @_;
    $param{config} ||= PureMessage::Scheduler::Config->new->as_hash;
    $param{logger} ||= PureMessage::Logger->new(ident => "scheduler");

    my $self = bless \%param, $class;

    # merge event_notify info into each event
    $self->_merge_notify_info($self->{config});
    
    # XXX the defaults should be put in the Scheduler::Config validator
    # (wait until AS::Config is sorted out for this release though)
    my $s_config = $self->{config};
    $self->{quit_file} = expand_filename(
	$self->{quit_file}   || 
	$s_config->{quit_file} || 
	"var/run/scheduler.quit");
    $self->{trigger_file} = expand_filename(
        $self->{trigger_file}   ||
        $s_config->{trigger_file} ||
        "var/run/scheduler.trig");
    my %default = (
	sleep_length => 1,
	run_limit => 0,
	run_interval => 0,
	use_utc_time => 0,
	verbose => 0,
    );
    foreach (keys %default) {
	$self->{$_} = $s_config->{$_} unless defined $self->{$_};
	$self->{$_} = $default{$_} unless defined $self->{$_};
    }

    $self->_fixup_env('PATH', "$PREFIX/bin");
    $self->_fixup_env('LD_LIBRARY_PATH', "$PREFIX/lib");

    return $self;
}

sub run {
    my $self = shift;

    my $s_config     = $self->{config};
    my $run_limit    = $self->{run_limit};
    my $sleep_length = $self->{sleep_length};
    my $run_interval = $self->{run_interval};
    my %c_events     = %{$s_config->{event}} if $s_config->{event};
    my $quit_file    = $self->{quit_file};
    my $trigger_file = $self->{trigger_file};

    $self->_dump_options() if $self->{verbose};

    return unless $s_config->{event};

    # XXX : last_run and last_dst should be stateful
    # (i.e. they should survive reboots of the OS & scheduler)
    my $last_run = time;
    my $last_dst = ($self->{use_utc_time})
	? (gmtime($last_run))[8]
	: (localtime($last_run))[8];

    my @events = $self->_prepare_events($last_run,%c_events);
    my %triggers = ();

    my $run_interval_count = 0;
    my $run_count = 0;

    # run continuously until the shutdown file is seen
    while (!$run_limit || ($run_count <= $run_limit)) {
	if (-e $quit_file) {
	    $self->_log("DEBUG", "Scheduler shutting down");
	    return $self->_shutdown;
	}
	$self->_log("DEBUG", "SCHED LOOP $run_count/$run_limit") 
	    if $self->{verbose};

        %triggers = (%triggers, $self->_read_trigger_file);

	if (!$run_interval || ++$run_interval_count == $run_interval) {
	    $run_interval_count = 0;

	    my $now = time;
	    my $now_dst = ($self->{use_utc_time})
		? (gmtime($now))[8]
		: (localtime($now))[8];

	    # Select the events that need to be run. Make sure that we pickup
	    # any events that may have been missed since the last run due to
	    # a mismatch in the run-interval.  Also, recompile events that
	    # either didn't update their next_run time or have a stale
	    # next_run due to a DST change.
	    my @run_events;
	    foreach my $event (@events) {
		next unless $event->ready;

		my $is_triggered = 0;
		if ($triggers{$event->{id}}) {
		    push @run_events, $event;
		    $is_triggered = 1;
		}

		next if $event->is_trigger_only;

		my $next_run = $event->next_run;
		if (!$next_run || 
		    $now_dst != $last_dst || 
		    ($next_run < $last_run && $next_run < $now)
		) {
		    # recompile event if it doesn't look right
		    $self->_log("DEBUG", "Recompiling event $event->{id}");
		    $event->compile($now);
		    $next_run = $event->next_run;
		}

		next if $is_triggered; # don't run an event twice

		if ($next_run > $last_run && 
		    $next_run <= $now
		) {
		    push @run_events, $event;
		}
	    }

	    $self->_log("DEBUG", "Running events at ".time2iso($now)) 
		if @run_events;
	    
	    # Actually run the events
	    foreach my $event (@run_events) {
		$self->_log("DEBUG", "Run the event $event->{id}");
		$event->run(id => $event->{id}, data => $event);
		delete $triggers{$event->{id}};
		unless ($event->is_trigger_only) {
		    $event->calc_next_run($now);
                    $self->_log("DEBUG", 
                        "event $event->{id} next run: " .
                        time2iso($event->next_run)) if $self->{verbose};
		}
	    }

	    $last_run = $now;
	    $last_dst = $now_dst;
	    $run_count++;
	}

	sleep($sleep_length);
    }
}

sub _read_trigger_file {
    my $self = shift;
    my $trigger_file = $self->{trigger_file};
    my $contents;
    
    # this isn't exactly atomic, but we can afford to wait
    return() unless(-f $trigger_file && -s _);
    $self->_log('DEBUG',"reading trigger file...");

    eval {
        my $tf = ActiveState::File::Atomic->new(
		$trigger_file, 
		writable => 1,
		timeout => 1, # second(s)
	    );
	$contents = $tf->slurp;
	$tf->commit_string(""); # clears and closes the file
    };
    if ($@) {
	$self->_log("WARNING","cannot read trigger file: $@");
	return();
    }

    $self->_log('DEBUG',"read trigger file");
    return() unless $contents;

    my %events = map {$_=>1} (grep(!/^$/, split(/\n/s, $contents)));

    if ($self->{verbose}) {
	$self->_log("DEBUG", "triggering events:");
	foreach (keys %events) {
	    $self->_log("DEBUG", "-> $_");
	}
    }

    return %events;
}

sub _prepare_events {
    my $self = shift;
    my $now = shift;
    my %c_events = @_;
    my $s_config = $self->{config};
    my $notification;

    # Job email notifications / Per-job logging
    eval {
		my $who  = $config{admin};
        $who =~ /<(.*)>/m;
        $who = $1;

        $notification = PureMessage::Scheduler::Notify->new(
             from_addr       => "$who",
             to_addr         => $config{admin},
             notify_on_error => _is_enabled($s_config->{notify}),
             logger          => $self->{logger},
             verbose         => $self->{verbose},
             event_logfile   => $s_config->{logfile},
             event_logging   => _is_enabled($s_config->{debug}),
        );
    };
    if ($@) {
	$self->_log("WARNING","Cannot enable notifications: $@");
    }

    # Sort the events so that they get initialized and run according to
    # their priority. Higher numbered events get a higher priority.
    my @event_keys = sort { 
	($c_events{$b}->{priority} || 0) <=> ($c_events{$a}->{priority} || 0)
    } keys %c_events;
    _dump_event_order(@event_keys) if $self->{verbose};
    
    # pre-load the class events, compile the 'when' and add sucessful
    # events to the %events hash.
    my @events;
    foreach my $id (@event_keys) {
	my $enabled = $c_events{$id}{enabled} || '';
	if (_is_enabled($enabled)) {
	    my $o = eval { PureMessage::Scheduler::Event->new(
		    id           => $id,
		    data         => $c_events{$id},
		    logger       => $self->{logger},
		    verbose      => $self->{verbose},
		    use_utc_time => $self->{use_utc_time},
		    init_time    => $now,
                    notification => $notification,
		)};
	    if ($@) {
		$self->_log("ERROR", "Failed to load event $id: $@");
		next;
	    }
	    push(@events, $o);
	    $self->_log("DEBUG", "Added event $id to event list");
	}
	else {
	    $self->_log("DEBUG", "Event $id is disabled");
	}
    }

    return @events;
}

sub _shutdown {
    my $self = shift;
    unlink($self->{quit_file}); # make sure this is gone
    return;
}

# prepend the component to the specified env variable, if it's not already
# there
sub _fixup_env {
    my ($self, $variable, $component) = @_;
    my $path = $ENV{$variable};
    my @path_dirs = ($path) ? split(':', $path) : ();
    unless (grep(/\A$component\Z/, @path_dirs)) {
        $self->_log (
            "WARNING",
            "Incorrect $variable environment variable; fixing by prepending $component"
        );
        unshift(@path_dirs, $component);
        $ENV{$variable} = join(':', @path_dirs);
    }
}

sub _log {
    my ($self, $level, $msg) = @_;
    print STDERR "$level: $msg\n" if $self->{verbose};
    $self->{logger}->log($level, $msg);
}

sub _dump_options {
    my $self = shift;

    print STDERR "Option: $_ = $self->{$_}\n"
	for qw(sleep_length run_interval quit_file);
}

sub _dump_event_order {
    print STDERR "Event sequence:\n";
    foreach (@_) {
	print STDERR "$_\n";
    }
}

sub _is_enabled {
    my $value = shift || '';
    return ($value =~ /^(?:1|y|yes)$/i) ? 1 : 0;
}

sub _merge_notify_info {
    my ($self, $config) = @_;

    return unless (exists $config->{event_notify});

    my $main_event = $config->{event};
    my $notify_info = $config->{event_notify}{event};
    for my $event_name (keys %$notify_info) {
        # merge
        if (exists $main_event->{$event_name}) {
            my $event_attributes = $notify_info->{$event_name};
            for my $item_key (keys %$event_attributes) {
                $main_event->{$event_name}{$item_key} = $event_attributes->{$item_key};
            }
        } else {
            my $msg = "Event ID '$event_name' in event-notify.conf does not exist.";
            $self->_log("WARNING", $msg);
            warn "$msg\n";
        }
    }
    delete $config->{event_notify};
}


1;
