package PureMessage::Scheduler::Config;
use strict;
use ActiveState::Config qw(:path);
use HTTP::Date qw(parse_date);
use Carp;
use PureMessage::Config qw(expand_config_filename);
use PureMessage::Scheduler::Event qw(%TIME_RANGE);

sub new {
    my $class = shift;
    my %args = (
	config_file => expand_config_filename("scheduler"),
	writable => 0,
	@_
    );

    my $cfg_file = delete($args{config_file});
    my $writable = delete($args{writable});
    my $self = bless {
	file        => $cfg_file,
	writable    => $writable,
	args        => \%args,
    }, ref($class) || $class;

    $self->refresh;

    return $self;
}

sub refresh {
    my $self = shift;
    require PureMessage::Config::Validator;

    # Revert previous changes
    delete $self->{cfg};

    my $p = ActiveState::Config->new(
	strict => 1,
	allow_includes => 1,
	%{$self->{args}},
    );
    PureMessage::Config::Validator->define_core_validators($p);
    PureMessage::Scheduler::Config->validate($p);

    $p->parse_file($self->{file}, writable => $self->{writable});

    # examine each node, and if it is an event with an action that looks like
    # a data update then alter the { m=*/freq, s=0 } entry to have the same
    # frequency of execution but a random offset in time.
    #
    my $iter = $p->root->children;
    while (my $node = $iter->next) {
        # only process if node is event node
        my $type = $node->groupclass() || '';
        next unless ($type eq 'event');

	# skip nodes without children since they can't be events
	my $node_iter = $node->children || next;

	# look for an event node with a data update action
	my $when_node;
	my $randomize = 0;
	while (my $event_node = $node_iter->next) {
	    my $name = $event_node->name;
	    my $value = $event_node->value;
	    my $groupclass = $event_node->groupclass;

	    if (
		defined($name) &&
		defined($value) && (
		    (($name eq 'action') &&
		     ($value =~ /ppm\b.*\bverify\b.*\bdata\b/i)) ||
		    (($name eq 'randomize') &&
		     ($value =~ /^\s*(?:1|y|yes)\s*$/i))
	        )
	       ){
		$randomize = 1;
	    }

	    if (defined($groupclass) && ($groupclass eq 'when')) {
		$when_node = $event_node;
	    }
	}

	if ($randomize && $when_node) {
	    my $when_iter = $when_node->children;
	    while (my $time_entry = $when_iter->next) {
		my $name = $time_entry->name;
	        my $value = $time_entry->value;
		next unless defined($name) && defined($value);

		if (exists $TIME_RANGE{$name}) {
		    my $tr = $TIME_RANGE{$name};
		    my $low = $tr->[0];
		    my $high = $tr->[$#$tr];

		    if ($value =~ /^\s*\*\/(\d+)\s*$/) {
			my $frequency = $1;
			my $new_value = int(rand($frequency)) + $low;
			my $i = $new_value + $frequency;
			while ($i <= $high) {
			    $new_value .= ",$i";
			    $i += $frequency;
			}
			$time_entry->value($new_value);
		    }
		    elsif ($value =~ /^\s*\d+\s*$/) {
			$time_entry->value(int(rand(($high-$low)+1))+$low);
		    }
		}
	    }
	}
    }
    $self->{cfg} = $p;
}

sub _valid_time {
    my ($p, $path, $node, $val, $append, $ctx, $ts) = @_;

    return if !defined($val);
    my @values = (ref $val eq 'ARRAY') ? @$val : ($val);
    @values = map {split ','} @values;
    
    my $type = $ts->[0]; 
    my $tr = $TIME_RANGE{$type};
    my $min = $tr->[0];
    my $max = $tr->[$#$tr];

    foreach my $value (@values) {
	next unless defined $value;
	$value =~ s/^\s+(.+?)\s+$/$1/; # trim whitespace
	next if $value eq '*';
	if ($value =~ /^\d+$/) {
	    # simple range check
	    die "value '$value' is too small (min=$min)" if $value < $min;
	    die "value '$value' is too large (max=$max)" if $value > $max;
	}
	elsif ($value =~ /^(\d+)-(\d+)$/) {
	    # range
	    my ($bot,$top) = ($1,$2);
	    die "top of range below bottom ".
	    "(perhaps you meant '$top\-$bot'?)"
		if ($top < $bot);
	    die "bottom '$bot' is too small" if $bot < $min;
	    die "bottom '$bot' is too large" if $bot > $max;
	    die "top '$top' is too small" if $top < $min;
	    die "top '$top' is too large" if $top > $max;
	}
	elsif ($value =~ /^\*\/(\d+)$/) {
	    # step
	    my $step = $1;
	    die "step cannot be zero in '$value'" if $step == 0;
	    die "step '$step' is too small" if $step < $min;
	    die "step '$step' is too large" if $step > $max;
	}
	else {
	    die "syntax error for time-part '$value'";
	}
    }
    1;
}

sub validate {
    my ($pkg, $p) = @_;

    my %conf = (
	'enabled'               => ['#atom nodups #bool', 1],
        'debug'                 => ['#atom nodups #bool', 0],
        'logfile'               => ['#atom nodups #str', "scheduler_log"],
	'notify'                => ['#atom nodups #bool', 1],

	'event#*.type'          => ['#atom nodups #enum exec class', undef],
	'event#*.desc'          => ['#atom nodups #str',  undef],
	'event#*.action'        => ['#atom nodups #str',  undef],
	'event#*.enabled'       => ['#atom nodups #bool', undef],
	'event#*.randomize'     => ['#atom nodups #bool', undef],
	'event#*.priority'      => ['#atom nodups #int',  0],
	'event#*.trigger_only'  => ['#atom nodups #bool', undef],
	#'event#*.when.*'        => ['#atom nodups #str',  0],
	'event_notify.event#*.count_threshold'    => ['#atom nodups #int', 1],
    );

    # define '#sched' type
    $p->typedecl( '#sched', \&_valid_time, undef);

    # add type checking for the when fields
    foreach my $t (keys %PureMessage::Scheduler::Event::TIME_RANGE) {
	$conf{"event#*.when.$t"} = ["#atom nodups #sched $t", undef];
    }

    for my $key (keys %conf) {
	my ($t, $d) = @{$conf{$key}};
	$p->typedecl($key, $t) if defined $t;
	$p->default($key, $d)  if defined $d;
    }
}

sub as_hash {
    my $self = shift;
    $self->{cfg}->as_hash;
}

sub save {
    my $self = shift;
    $self->{cfg}->save;
}

sub root {
    my $self = shift;
    $self->{cfg}->root;
}

sub enable_event {
	my($self, $event) = @_;
	my $node;
	croak qq{Event "$event" not found}
	    unless($node = path_node($self->{cfg}->root, "event#$event"));
	path_value($node, "enabled", 1);
}

sub disable_event {
	my($self, $event) = @_;
	my $node;
	croak qq{Event "$event" not found}
	    unless($node = path_node($self->{cfg}->root, "event#$event"));
	path_value($node, "enabled", 0);
}

sub is_enabled_event {
    my ($self, $event) = @_;
    my $node;
    croak qq{Event "$event" not found}
        unless($node = path_node($self->{cfg}->root, "event#$event"));
    return path_value($node, "enabled");
}

sub delete_event {
    my ($self, $event) = @_;
    path_node($self->{cfg}->root, "event#$event", delete => 1);
}

sub run_event {
    my ($self, $event) = @_;
    $event->run(id => $event->{id}, data => $event);
}

# returns string representation of the "when" component of an event
sub schedule_to_string {
    my ($self, $schedule) = @_;
    my @sched_component;
    for my $key (keys %TIME_RANGE) {
        push @sched_component, "$key => $schedule->{$key}" 
            if (exists $schedule->{$key});
    }
    return join "; ", @sched_component;
}

1;
