package PureMessage::Service::scheduler;
use strict;
use base 'PureMessage::Service';
use Carp qw(croak);
use ActiveState::Handy qw(run);
use PureMessage::Config qw(expand_filename expand_config_filename);
use ActiveState::Config::Depend qw(config_mtime_deps);
use ActiveState::Unix::ProcInfo qw(proc_info);
use ActiveState::File::Atomic;
use PureMessage::_sys;
use Fcntl qw(:DEFAULT :flock);
use Errno qw(ESRCH);

my $schedule_d = expand_filename("etc/init.d/pmx-scheduler");
my $pid_file = expand_filename("var/run/schedule.pid");

sub Available {
    my $class = shift;
    return ($class->new(name => "Scheduler"));
}

sub start {
    my ($self, %opts) = @_;
    _exec([$schedule_d, "start"], %opts);
}

sub stop {
    my ($self, %opts) = @_;
    _exec([$schedule_d, "stop"], %opts);
}

sub restart {
    my ($self, %opts) = @_;
    _exec([$schedule_d, "restart"], %opts);
}

sub status {
    my %s = (running => 0);
    my $pid = eval { check_pid() };
    if ($pid) {
	$s{pid} = $pid;
	$s{running} = 1;
	$s{ready} = 1;
	my $pid_mtime = (stat($pid_file))[9];
	my $cf_mtime = config_mtime_deps(
		expand_config_filename("scheduler")
	    );
	$s{config} = $pid_mtime >= $cf_mtime ? "sync" : "stale";
    }
    return \%s;
}

sub running {
    return status()->{running};
}

sub trigger {
    my($self, %opts) = @_;

    my $event = $opts{event};
    croak "no trigger event specified" unless $event;

    my $trigger_file = expand_filename(
        $opts{trigger_file}     ||
        $self->{trigger_file}   ||
        "var/run/scheduler.trig"
    );
    
    eval {
        my $tf = ActiveState::File::Atomic->new(
		$trigger_file, 
		writable => 1,
		create => 1,
		timeout => 2, # seconds
	    );
	my $contents = $tf->slurp;
	$contents .= "$event\n";
        $tf->commit_string($contents);
    };
    if ($@) {
	croak "cannot write event '$event' to trigger file: $@";
    }
    return 1;
}

sub _exec {
    my $cmd = shift;
    my %opts = @_;

    if ($opts{bg}) {
	my $pid = fork;
	die "Can't fork: $!" unless defined $pid;
	return $pid if $pid;
	open(STDIN, "</dev/null");
	open(STDOUT, ">/dev/null");
	open(STDERR, ">/dev/null");
	exec(@$cmd);
	die "Failed to exec @$cmd: $!";
    }
    $$cmd[0] = '@-' . $$cmd[0];
    run(@$cmd);
}

# returns 0 if the scheduler is still running
# returns 1 if the pid file was successfully written
# dies on error
sub write_pid {
    my $pid_fh;

    return 0 if check_pid(\$pid_fh);

    my $info = proc_info(root_pid => $$);
    die "Cannot find my pid!!\n" unless $info && $info->{$$};
    my $cmd = $info->{$$}{args};

    sysseek($pid_fh,0,0);
    truncate($pid_fh,0);
    my $pidstr = "$$ $cmd\n";
    my $l = length($pidstr);
    my $r = syswrite($pid_fh, $pidstr, $l);
    die "Cannot write to $pid_fh: $!"
	unless (defined($r) && $r==$l);

    close($pid_fh) or
	die "Cannot write to $pid_fh: $!";
    return 1;
}

# Returns true if the scheduler is running
sub check_pid {
    my $pid_fh_ref = shift;

    my ($pid, $cmd) = read_pid($pid_fh_ref);
    if ($pid) {
	if(kill(0,$pid)) {
            my $info = proc_info(root_pid => $pid);
            if ($info && $info->{$pid} && $info->{$pid}{args} eq $cmd) {
                return $pid;
            }
	}
    }
    return 0;
}

sub read_pid {
    my $pid_fh_ref = shift;

    my $pid_fh;
    sysopen($pid_fh, $pid_file, O_RDWR|O_CREAT)
	or die "Cannot open pid file '$pid_file': $!";
    flock($pid_fh, LOCK_EX|LOCK_NB)
	or die "Cannot lock pid file '$pid_file' in EX mode: $!";

    # now that we have the lock, we can reliably check the pid
    my ($pid, $cmd) = <$pid_fh> =~ /^(\d+)\s(.*)/;
    $pid ||= 0;
    $cmd ||= '';

    if ($pid_fh_ref) {
	$$pid_fh_ref = $pid_fh;
    } else {
	close $pid_fh;
    }

    return wantarray ? ($pid, $cmd) : $pid;
}

sub unlink_pid {
    my %args = @_;
    my $pid = read_pid();
    if ($args{force}) {
        unlink $pid_file or warn "Can't unlink pid file '$pid_file': $!";
    } else {
        if ($pid == $$) {
            unlink $pid_file or warn "Can't unlink pid file '$pid_file': $!";
        } 
    }
}

1;
